#include "stdio.h"
#include "stdlib.h"
#include "math.h"

#include "Algorithme2.h"


/* --------------------------------------------------------------------------- *
 * Ce programme est l'implémentation de l'algorithme heuristique. La fonction
 * communitiesPartition prend comme argument un tableau d'entiers à deux
 * dimensions qui correspond à la matrice d'adjacence du graphe, elle prend
 * également comme argument un entier n qui correspond à la taille de la matrice
 * d'adjacence. Lorsque l'algorithme est terminé, la fonction écrit dans un
 * fichier nommé "meilleureCom.txt" les différentes communautés détectées dans
 * le graphe.
 * -------------------------------------------------------------------------- */




static EnsembleDesCommunautes creationCommunautes(int n) {
    int i;

    EnsembleDesCommunautes newEnsembleCom;
    newEnsembleCom.nombreCommunautes = n;

    Communaute *vecteurCom;
    vecteurCom = (Communaute*)malloc(n*sizeof(Communaute));
	if(vecteurCom==NULL){
		printf("Allocation impossible\n");
		exit(1);
	}

    for(i=0; i<n; i++){
        Communaute Com;
        Com.numero = i;
        Com.nbrElementsActuel = 1;
        Com.nbrElementsDepart = 1;
        int* sommetsActuels = (int*)malloc(1*sizeof(int));
        sommetsActuels[0] = i;
        Com.listeSommetsActuel = sommetsActuels;
        int* sommetsDepart = (int*)malloc(1*sizeof(int));
        sommetsDepart[0] = i;
        Com.listeSommetsDepart = sommetsDepart;
        vecteurCom[i] = Com;
    }
    newEnsembleCom.vecteurCommunautes = vecteurCom;
    return newEnsembleCom;
}

static void ecritureEnsComDansFichier(EnsembleDesCommunautes meilleureCom, char* nomDeFichier){
    size_t i,j;
    FILE* fichierEcriture = NULL;
    fichierEcriture = fopen(nomDeFichier,"w");
    if(fichierEcriture==NULL){
        printf("Impossible d'ouvrir le fichier \n");
        exit(1);
    }

    fprintf(fichierEcriture, "nombre de communautes = %zu \n", meilleureCom.nombreCommunautes);
    for(i=0; i<meilleureCom.nombreCommunautes; i++){
        fprintf(fichierEcriture, "com %zu = numero - %zu, nombreElements - %zu, listeSommets - ", i, meilleureCom.vecteurCommunautes[i].numero, meilleureCom.vecteurCommunautes[i].nbrElementsDepart);
        for(j=0; j<meilleureCom.vecteurCommunautes[i].nbrElementsDepart; j++){
            fprintf(fichierEcriture, "%d - ", meilleureCom.vecteurCommunautes[i].listeSommetsDepart[j]);
        }
        fprintf(fichierEcriture, "\n");
    }

    fclose(fichierEcriture);
}

static int communauteAvecUnElement(EnsembleDesCommunautes com, int indexDepart){
    size_t nbrCom = com.nombreCommunautes;
    size_t i = indexDepart;
    int sommet = 0;
    while(sommet==0 && i<nbrCom){
        if(com.vecteurCommunautes[i].nbrElementsActuel == 1){
            sommet = 1;
        }
        else {
            i++;
        }
    }
    if(sommet == 0){
        return -1;
    }
    else{
        return i;
    }
}

static int sommeElementsNonNuls(int** matrice, int n){
    int m = 0;
    int i,j;
    for(i=0; i<n; i++){
        for(j=0; j<n; j++){
            if(i!=j){
                m = m + matrice[i][j];
            }
        }
    }
    m = m/2.0;
    for(i=0; i<n; i++){
        m = m + matrice[i][i];
    }
    return m;
}

static int communauteAdjacente(int** matriceAdjacence, Communaute com1, Communaute com2){
    size_t i,j;
    int adjacent = 0;
    for(i=0; i<com1.nbrElementsActuel; i++){
        for(j=0; j<com2.nbrElementsActuel; j++){
            if(adjacent==0){
                if(matriceAdjacence[com1.listeSommetsActuel[i]][com2.listeSommetsActuel[j]] != 0){
                    adjacent = 1;
                }
            }
        }
    }
    return adjacent;
}

static int nombreArreteEntreSommetEtCom(Communaute com, int numSommetLabel, int** matriceAdjacence){
    int nbrArretes = 0;
    int nbrElements = com.nbrElementsActuel;
    int i;
    int* tabLabelSommets = (int*) malloc(nbrElements*sizeof(int));
    for(i=0; i<nbrElements; i++){
      tabLabelSommets[i] = com.listeSommetsActuel[i];
    }
    for(i=0; i<nbrElements; i++){
            nbrArretes = nbrArretes + matriceAdjacence[numSommetLabel][tabLabelSommets[i]];
    }
    return nbrArretes;
}

static int nombreArreteVersCommunaute(Communaute com, int** matriceAdjacence, int n){
    int nbrArretes = 0;
    int nbrElements = com.nbrElementsActuel;
    int i,j;

    int* tabLabelSommets = (int*) malloc(nbrElements*sizeof(int));
    for(i=0; i<nbrElements; i++){
      tabLabelSommets[i] = com.listeSommetsActuel[i];
    }

    for(i=0; i<nbrElements; i++){
        for(j=0; j<n; j++){
            nbrArretes = nbrArretes + matriceAdjacence[tabLabelSommets[i]][j];
        }
    }
    for(i=0; i<nbrElements; i++){
        for(j=i+1;j<nbrElements;j++){
            nbrArretes = nbrArretes - (2*matriceAdjacence[tabLabelSommets[i]][tabLabelSommets[j]]);
        }
    }
    return nbrArretes;
}

static EnsembleDesCommunautes fusionCommunautes(EnsembleDesCommunautes ensembleCommunautes, Communaute com1, Communaute com2){
    size_t i;
    Communaute newCommunaute;

    size_t nbrActuel = com1.nbrElementsActuel + com2.nbrElementsActuel;
    newCommunaute.nbrElementsActuel = nbrActuel;

    size_t nbrDepart = com1.nbrElementsDepart + com2.nbrElementsDepart;
    newCommunaute.nbrElementsDepart = nbrDepart;

    int* sommetsActuels = (int*) malloc(nbrActuel*sizeof(int));
    for(i=0; i<com1.nbrElementsActuel; i++){
        sommetsActuels[i] = com1.listeSommetsActuel[i];
    }
    for(i=com1.nbrElementsActuel; i<nbrActuel; i++){
        sommetsActuels[i] = com2.listeSommetsActuel[i-com1.nbrElementsActuel];
    }
    newCommunaute.listeSommetsActuel = sommetsActuels;

    int* sommetsDepart = (int*) malloc(nbrDepart*sizeof(int));
    for(i=0; i<com1.nbrElementsDepart; i++){
        sommetsDepart[i] = com1.listeSommetsDepart[i];
    }
    for(i=com1.nbrElementsDepart; i<nbrDepart; i++){
        sommetsDepart[i] = com2.listeSommetsDepart[i-com1.nbrElementsDepart];
    }
    newCommunaute.listeSommetsDepart = sommetsDepart;

    int numPetit;
    int numGrand;
    if(com1.numero<com2.numero){
        numPetit = com1.numero;
        numGrand = com2.numero;
    }
    else{
        numPetit = com2.numero;
        numGrand = com1.numero;
    }
    newCommunaute.numero = numPetit;
    ensembleCommunautes.vecteurCommunautes[numPetit] = newCommunaute;
    for(i=numGrand; i<ensembleCommunautes.nombreCommunautes-1; i++){
        ensembleCommunautes.vecteurCommunautes[i] = ensembleCommunautes.vecteurCommunautes[i+1];
        ensembleCommunautes.vecteurCommunautes[i].numero = ensembleCommunautes.vecteurCommunautes[i].numero - 1;
    }
    ensembleCommunautes.nombreCommunautes = ensembleCommunautes.nombreCommunautes - 1;
    return ensembleCommunautes;
}

static int possibleAssocierSommetCommunaute(EnsembleDesCommunautes com, int numSommetCom, int** matriceAdjacence, int n, int m){
    int nbrCom = com.nombreCommunautes;
    int i,k;
    int max = -1;
    int positionMax;
    int nbrArreteVersSommet = 0;
    int labelSommet = com.vecteurCommunautes[numSommetCom].listeSommetsActuel[0];
    for(k=0; k<n; k++){
        nbrArreteVersSommet = nbrArreteVersSommet + matriceAdjacence[labelSommet][k];
    }
    for(i=0; i<nbrCom; i++){
        if(i!=numSommetCom){
            int adj = communauteAdjacente(matriceAdjacence, com.vecteurCommunautes[numSommetCom], com.vecteurCommunautes[i]);
            if(adj == 1){
                int tot;
                int nbrLien = nombreArreteEntreSommetEtCom(com.vecteurCommunautes[i], com.vecteurCommunautes[numSommetCom].listeSommetsActuel[0], matriceAdjacence);
                int nbrArreteVersCom = nombreArreteVersCommunaute(com.vecteurCommunautes[i], matriceAdjacence, n);
                tot = (nbrLien*m) - (nbrArreteVersSommet*nbrArreteVersCom);
                if(tot>max){
                    max = tot;
                    positionMax = i;
                }
            }
        }
    }

    if(max<0){
        return -1;
    }
    else{
        return positionMax;
    }
}

static EnsembleDesCommunautes algoMettreSommetsEnsemble(EnsembleDesCommunautes com, int** matriceAdjacence, int n){

    int m = sommeElementsNonNuls(matriceAdjacence, n);
    size_t indexDepart = 0;
    int okContinue = communauteAvecUnElement(com, indexDepart);

    while(okContinue != -1){
        int ok = possibleAssocierSommetCommunaute(com, okContinue, matriceAdjacence, n, m);
        if(ok == -1){
            okContinue = -1;
        }
        else{
            com = fusionCommunautes(com, com.vecteurCommunautes[okContinue], com.vecteurCommunautes[ok]);
            okContinue = communauteAvecUnElement(com,0);
        }

    }
    return com;
}

static int sommetLiensEntreDeuxCom(int** matriceAdjacence, Communaute com1, Communaute com2){

    int nbr1 = com1.nbrElementsActuel;
    int nbr2 = com2.nbrElementsActuel;
    int s = 0;
    int i,j;
    for(i=0; i<nbr1; i++){
        for(j=0; j<nbr2; j++){
            s = s + matriceAdjacence[com1.listeSommetsActuel[i]][com2.listeSommetsActuel[j]];
        }
    }
    return s;
}

static int sommetLiensInterCom(int** matriceAdjacence, Communaute com){
    int i,j;
    int s = 0;
    int nbrElement = com.nbrElementsActuel;
    for(i=0; i<nbrElement; i++){
        for(j=0; j<nbrElement; j++){
            s = s + matriceAdjacence[com.listeSommetsActuel[i]][com.listeSommetsActuel[j]];
        }
    }
    return s;
}

static int** nouvelleMatriceAdjacence(EnsembleDesCommunautes com, int** matriceAdjacence, int n){
    int newTaille = com.nombreCommunautes;
    int i,j;
    int ** newMatriceAdjacence = (int**)malloc(newTaille*sizeof(int*));
    if(newMatriceAdjacence==NULL){
        printf("Allocation impossible\n");
        exit(1);
    }

    for(i=0; i<newTaille; i++){
        newMatriceAdjacence[i]=(int*)malloc(n*sizeof(int));
        if(newMatriceAdjacence[i]==NULL){
            printf("Allocation impossible\n");
            exit(1);
        }
    }

    for(i=0; i<newTaille; i++){
        for(j=0; j<newTaille; j++){
            if(i==j){
                int s = sommetLiensInterCom(matriceAdjacence, com.vecteurCommunautes[i]);
                newMatriceAdjacence[i][j]=s;
            }
            else{
                int s = sommetLiensEntreDeuxCom(matriceAdjacence, com.vecteurCommunautes[i], com.vecteurCommunautes[j]);
                newMatriceAdjacence[i][j]=s;
            }
        }
    }
    return newMatriceAdjacence;
}

static EnsembleDesCommunautes reinitialiserCom(EnsembleDesCommunautes com){
    int n = com.nombreCommunautes;
    int i;
    for(i=0;i<n;i++){
      com.vecteurCommunautes[i].nbrElementsActuel = 1;
      int* sommetsActuels = (int*)malloc(1*sizeof(int));
      sommetsActuels[0] = i;
      com.vecteurCommunautes[i].listeSommetsActuel = sommetsActuels;
    }
    return com;
}

void communitiesPartition(int **matriceAdjacence, int n){
    EnsembleDesCommunautes com = creationCommunautes(n);
    int okContinue = 1;
    while(okContinue==1){
        com = algoMettreSommetsEnsemble(com, matriceAdjacence, n);
        matriceAdjacence = nouvelleMatriceAdjacence(com, matriceAdjacence, n);
        int nbr = com.nombreCommunautes;
        if(nbr<n){
            okContinue = 1;
        }
        else{
            okContinue = 0;
        }
        n = nbr;
        com = reinitialiserCom(com);
    }
    ecritureEnsComDansFichier(com, "meilleureCom.txt");
}
